/**************************************************************************
   watsup    Linux system and process monitoring program

   Copyright 2007-2014 Michael Cornelison
   source URL:  kornelix.com
   contact: kornelix@posteo.de
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************

   main program
   
   $ watsup [ -delay NN ] [ -proc ] [ -disk /dev/sda ]

   -delay NN         optional startup delay in seconds
   -proc             show top process resources details
   -disk /dev/sdx    disk to show temperature
   
   Tiny window is shown with CPU load, disk IO rate, network IO rate.
   If -proc, process window is shown with top process resource details.

***************************************************************************/

#include <dirent.h>
#include <sys/resource.h>
#include "zfuncs.h"

#define Tinterval 1000                                                     //  tiny window update period, millisecs
#define Ptitle "watsup v.3.0"                                              //  process window title

int   Twin_update(void *);                                                 //  periodic update tiny window
int   Twin_mouse(GtkWidget *, GdkEventButton *);                           //  tiny window mouse event
void  getloads();                                                          //  get CPU, disk, network stats
void  gettemps();                                                          //  get core temperature range
int   Pmain();                                                             //  show top process details
void  load_config();                                                       //  load config parameters
void  save_config();                                                       //  save config parameters

int         m_argc;                                                        //  argc and argv[] parameters
char        **m_argv;
char        configfile[200];                                               //  /home/<user>/.watsup/config
double      cpuload;                                                       //  total CPU load (100% = 1 core)
double      diskrate;                                                      //  disk IO rate, MB/sec
double      netrate;                                                       //  network IO rate, KB/sec
char        disk[20];                                                      //  "/dev/sda" etc. for disk temp.
int         coreloT, corehiT;                                              //  low and high core temperatures
int         diskT = 0;                                                     //  disk temperature
int         Pstart, Pruns;                                                 //  flags, control process report
int         txpos = 200, typos = 100;                                      //  default tiny window position
int         tww = 200, thh = 25;                                           //  default tiny window size
int         pxpos = 200, pypos = 150;                                      //  default process window position
int         pww = 500, phh = 400;                                          //  default process window size
char        matchhist[20][30];                                             //  history of match strings used
int         Nmatchhist = 0;                                                //  count of match strings <= 20

GtkWidget      *Twin;                                                      //  tiny main window
GtkWidget      *Twinlabel;


int main(int argc, char *argv[])                                           //  tiny window startup
{
   m_argc = argc;                                                          //  save for beroot() call    v.2.7
   m_argv = argv;
   
   int      delay;                                                         //  startup delay
   cchar    *startlab = " Right click exit. Left click top processes ";
   
   delay = 2;                                                              //  default delay
   strcpy(disk,"/dev/sda");                                                //  default disk drive
   
   gtk_init(&argc,&argv);                                                  //  initz. GTK

   zinitapp("watsup",null);                                                //  get app directories
   
   for (int ii = 1; ii < argc; ii++)                                       //  get command line params
   {
      if (strEqu(argv[ii],"-delay")) {
         if (argc < ii+2) continue;
         delay = atoi(argv[ii+1]);
         if (delay < 0 || delay > 99) {
            printf("delay set to 5 \n");
            delay = 5;
         }
         sleep(delay);                                                     //  startup delay
         ii++;
      }
      
      else if (strEqu(argv[ii],"-proc")) Pstart = 1;                       //  start top processes window
      
      else if (strEqu(argv[ii],"-disk")) {                                 //  disk: /dev/sda etc.
         if (argc < ii+2) continue;
         strncpy0(disk,argv[ii+1],20);
         ii++;
      }
      
      else {
         printf("%s \n",argv[ii]);
         printf("usage: $ watsup [ -delay NN ] [ -proc ] [ -disk /dev/sdx ] \n");
         return 1;
      }
   }
   
   load_config();                                                          //  get configuration parameters

   Twin = gtk_window_new(GTK_WINDOW_POPUP);                                //  main (tiny) window
   gtk_window_set_default_size(GTK_WINDOW(Twin),tww,thh);                  //  set size and position
   gtk_window_move(GTK_WINDOW(Twin),txpos,typos);

   Twinlabel = gtk_label_new(startlab);                                    //  label for data display
   gtk_container_add(GTK_CONTAINER(Twin),Twinlabel);

   gtk_widget_add_events(Twin,GDK_BUTTON_PRESS_MASK);                      //  connect mouse events
   gtk_widget_add_events(Twin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(Twin,GDK_BUTTON_MOTION_MASK);                     //  any button down + mouse movement
   G_SIGNAL(Twin,"button-press-event",Twin_mouse,0);
   G_SIGNAL(Twin,"button-release-event",Twin_mouse,0);
   G_SIGNAL(Twin,"motion-notify-event",Twin_mouse,0);

   g_timeout_add(Tinterval,Twin_update,0);                                 //  start periodic update function

   gtk_widget_show_all(Twin);                                              //  show all widgets

   gtk_main();                                                             //  process GUI events
   return 0;
}


//  periodic function to retrieve and display loadings and temperatures

int Twin_update(void *)
{
   static int  ftf = 3;
   char        text[100];

   getloads();                                                             //  loadings
   gettemps();                                                             //  temperatures
   
   if (ftf) {                                                              //  discard first samples
      ftf--;
      return 1;
   }
                                                                           //  output panel report
   sprintf(text," CPU %03.0f%c  %dC   disk %.0f MB/s  %dC   net %.0f KB/s ",
                     cpuload,'%',corehiT,diskrate,diskT,netrate);          //  v.2.7
   gtk_label_set_text(GTK_LABEL(Twinlabel),text);
   
   if (Pstart) {                                                           //  start process detail report
      Pstart = 0;
      if (! Pruns) Pmain();
   }
   
   return 1;
}


//  respond to mouse drag or click:
//  move tiny window, kill tiny window, start top process report window

int Twin_mouse(GtkWidget *, GdkEventButton *event)                         //  overhauled  v.2.0
{
   static int     Fdrag = 0;                                               //  drag motion underway
   static int     mposx0, mposy0;
   int            mposx, mposy;
   
   if (event->type == GDK_MOTION_NOTIFY)                                   //  drag underway
   {
      mposx = event->x;                                                    //  curr. mouse position
      mposy = event->y;
   
      if (! Fdrag) {                                                       //  new drag was started
         Fdrag = 1;
         mposx0 = mposx;                                                   //  drag start position
         mposy0 = mposy;
         return 1;
      }

      gtk_window_get_position(GTK_WINDOW(Twin),&txpos,&typos);             //  window position
      txpos += mposx - mposx0;                                             //  + drag motion
      typos += mposy - mposy0;                                             //  = new window position
      gtk_window_move(GTK_WINDOW(Twin),txpos,typos);                       //  (relative mouse position unchanged)
   }
   
   if (event->type == GDK_BUTTON_RELEASE)                                  //  mouse drag or click finished
   {
      if (Fdrag) {                                                         //  drag was ended
         Fdrag = 0;
         save_config();                                                    //  save configuration parameters
         return 1;
      }
      
      if (event->button == 1)                                              //  left button
         Pstart = 1;                                                       //  start process detail report
      
      if (event->button == 3) {                                            //  right button - quit
         save_config();                                                    //  save configuration parameters
         gtk_main_quit();                                                  //  exit application
      }
   }

   return 1;
}


//  Return CPU load in percent, disk IO in MB/sec, network IO in KB/sec.
//  Statistics are based on the interval between prior and current calls.
//  CPU load: 100 corresponds to one processor core fully loaded.
//  Zeros are returned for the first call.

void getloads()
{
   static int     ftf = 1;                                                 //  1st time flag
   FILE           *fid;
   char           *pp, buff[1000];
   const char     *ppc;
   double         val1, val2;
   static double  jifftime;                                                //  CPU time slice
   static double  time1, time2, elapsed;
   static double  jiff1, jiff2, dio1, dio2, net1, net2;

   if (ftf) jifftime = 1.0 / sysconf(_SC_CLK_TCK);                         //  "jiffy" time slice = 1.0 / HZ
   
   jiff2 = dio2 = net2 = 0;

   fid = fopen("/proc/stat","r");                                          //  read /proc/stat
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         if (strncmp(pp,"cpu ",4) == 0) {                                  //  look for "cpu " record (all cores)
            parseprocrec(pp,2,&val1,4,&val2,null);                         //  get user and system mode jiffies
            jiff2 += val1 + val2;
         }
      }
      fclose(fid);
   }
   
   fid = fopen("/proc/diskstats","r");                                     //  read disk statistics
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         ppc = strField(buff,' ',1);                                       //  screen for "8"
         if (! ppc) continue;
         if (strNeq(ppc,"8")) continue;
         ppc = strField(buff,' ',3);                                       //  disk, "sda" or "sda1" etc.
         if (! ppc) continue;
         if (ppc[3]) continue;                                             //  ignore partitions
         parseprocrec(pp,6,&val1,10,&val2,null);                           //  get sectors read and written
         dio2 += val1 + val2;
      }
      fclose(fid);
   }
   
   fid = fopen("/proc/net/dev","r");                                       //  read network statistics
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         pp = strchr(buff,':');                                            //  look for device "eth0:" etc.
         if (! pp) continue;
         parseprocrec(pp+1,1,&val1,9,&val2,null);                          //  get bytes received and transmitted
         net2 += val1 + val2;
      }
      fclose(fid);
   }
   
   time2 = get_seconds();                                                  //  get exact time since last sample
   
   if (ftf) {
      ftf = 0;
      time1 = time2;
      jiff1 = jiff2;
      dio1 = dio2;
      net1 = net2;
      cpuload = diskrate = netrate = 0;
      return;
   }
   
   elapsed = time2 - time1;
   cpuload = (jiff2 - jiff1) * jifftime / elapsed * 100.0;                 //  percent CPU time
   diskrate = (dio2 - dio1) * 512.0 / 1048576.0 / elapsed;                 //  disk megabytes / sec.
   netrate = (net2 - net1) / 1024.0 / elapsed;                             //  network kilobytes /sec.
   
   time1 = time2;                                                          //  save current counters for next call
   jiff1 = jiff2;
   dio1 = dio2;
   net1 = net2;

   return;   
}


//  read processor core temperatures and disk temperature

void gettemps()                                                            //  v.3.0
{
   static int  cycle = 8;

   coretemps(coreloT,corehiT);                                             //  get cpu temp. every cycle
   if (++cycle % 10 == 0) diskT = disktemp(disk);                          //  get disk temp. every 10th cycle
   return;
}


/**************************************************************************

   Report overal system loads and top process loads in a larger window

***************************************************************************/

//  Linux system parameters
int      pagesize;                                                         //  memory page size
double   jiffy;                                                            //  CPU time slice
int      ncpus;                                                            //  CPU count

//  program parameters and limits
#define  maxpid 1000                                                       //  max. PIDs sampled
#define  maxtop 100                                                        //  max. top PIDs to report
#define  proc_ignore "getty init"                                          //  processes to ignore
#define  mega (1024 * 1024)                                                //  computer megabyte
#define  autoscroll GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC              //  scroll window policy
#define  textwin GTK_TEXT_WINDOW_TEXT                                      //  GDK window of GTK text view

//  editable parameters and default values
double   Pinterval = 1.0;                                                  //  sample interval, 1 sec.
int      fontsize = 8;                                                     //  report font size
char     headerfont[20] = "Monospace bold 8";                              //  report header font
char     bodyfont[20] = "Monospace 8";                                     //  report body font
char     procmatch[100] = "";                                              //  process match string(s)

//  calculated parameters
int      fontheight = 20;                                                  //  report font height
int      topmost = 20;                                                     //  topmost procs within window
int      sysreplines = 8;                                                  //  report lines needed for system data
int      phh2;                                                             //  height of text window 

GtkWidget      *Pwin, *mVbox, *mScroll, *mLog;                             //  main window widgets

struct   sysrec {                                                          //  system statistics record layout
   double   cpubusy;
   double   memused, memfree, memcache;
   double   pfsoft, pfhard;
   double   diskops, disksecs;
   double   netrecv, netxmit;
};

sysrec      srec1, srec2;                                                  //  system stats, sample 1 and sample 2

struct   pidrec {                                                          //  process statistics record layout
   int      pid;
   int      row;
   double   weight;
   double   cpu;
   double   mem;
   double   pfs;
   double   dmbs;
   char     username[20];
   char     exefname[40];
};

pidrec      prec1[maxpid];                                                 //  proc recs, sample 1
pidrec      prec2[maxpid];                                                 //  proc recs, sample 2
pidrec      prec3[maxpid];                                                 //  proc recs for report
int         nrec1, nrec2, nrec3;                                           //  record counts for precN arrays
int         rec1, rec2, rec3;                                              //  indexes for precN arrays
int         repline;                                                       //  tracks report output lines

double      time1, time2;                                                  //  sample times 1 and 2
double      elapsed;                                                       //  elapsed time between 1 and 2
int         Fresort = 0;                                                   //  flag, re-sort topmost by weight

char        *matchstrings[10];                                             //  process match strings (wild)

GdkCursor   *cursor;
GdkWindow   *gdkwin;

int m_draw(GtkWidget *, cairo_t *cr);                                      //  main window draw event
void m_mouse(GtkWidget *, GdkEventButton *event, void *);                  //  main window mouse event

void menufunc(GtkWidget *item, const char *menu);                          //  menu functions
void m_parms();
void m_getroot();
void m_start();
void m_stop();
void m_sort();
void m_save();
void m_print();
void m_quit();
void m_help();

int  proc_report(void *);                                                  //  main process report

void get_sys_stats(sysrec *srec);                                          //  functions for system statistics
void report_sys_stats();
void copy_sys_stats();

void get_proc_stats(pidrec *prec, int &nnp);                               //  functions for process statistics
void report_proc_stats();
void copy_proc_stats();

void setreportfonts();                                                     //  change report fonts


int Pmain()
{
   GtkWidget   *tbar;
   int         err, prio;

   *procmatch = 0;
   matchstrings[0] = 0;

   Pwin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  process window
   gtk_window_set_title(GTK_WINDOW(Pwin),Ptitle);
   gtk_window_set_default_size(GTK_WINDOW(Pwin),pww,phh);                  //  set size and location
   gtk_window_move(GTK_WINDOW(Pwin),pxpos,pypos);

   mVbox = gtk_box_new(VERTICAL,0);                                        //  vertical packing box
   gtk_container_add(GTK_CONTAINER(Pwin),mVbox);                           //  add to main window

   tbar = create_toolbar(mVbox);                                           //  create tool bar and buttons
   add_toolbar_button(tbar,"parms","edit parameters","preferences.png",menufunc);
   add_toolbar_button(tbar,"root","become root","root.png",menufunc);
   add_toolbar_button(tbar,"start","start reporting","run.png",menufunc);
   add_toolbar_button(tbar,"stop","stop reporting","stop.png",menufunc);
   add_toolbar_button(tbar,"sort","resort top processes","refresh.png",menufunc);
   add_toolbar_button(tbar,"save","save report to a file","save.png",menufunc);
   add_toolbar_button(tbar,"print","print report","print.png",menufunc);
   add_toolbar_button(tbar,"quit","quit watsup","quit.png",menufunc);
   add_toolbar_button(tbar,"help","view user guide","help.png",menufunc);

   mScroll = gtk_scrolled_window_new(0,0);                                 //  scrolled window
   gtk_box_pack_start(GTK_BOX(mVbox),mScroll,1,1,0);                       //  add to main window mVbox
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(mScroll),autoscroll);

   mLog = gtk_text_view_new();                                             //  text window
   gtk_text_view_set_editable(GTK_TEXT_VIEW(mLog),0);
   gtk_container_add(GTK_CONTAINER(mScroll),mLog);                         //  add to scrolled window
   
   G_SIGNAL(Pwin,"destroy",m_quit,0);                                      //  connect window destroy event
   G_SIGNAL(Pwin,"draw",m_draw,0);                                         //  gtk3

   gtk_widget_add_events(mLog,GDK_BUTTON_RELEASE_MASK);                    //  connect mouse clicks  v.1.2
   G_SIGNAL(mLog,"button-release-event",m_mouse,0);
   gtk_widget_add_events(mLog,GDK_POINTER_MOTION_MASK);                    //  v.2.4
   G_SIGNAL(mLog,"motion-notify-event",m_mouse,0);

   gtk_widget_show_all(Pwin);                                              //  show all widgets

   cursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);                            //  use arrow cursor for text window
   gdkwin = gtk_text_view_get_window(GTK_TEXT_VIEW(mLog),textwin);
   gdk_window_set_cursor(gdkwin,cursor);

   jiffy = 1.0 / sysconf(_SC_CLK_TCK);                                     //  "jiffy" time slice = 1.0 / HZ
   pagesize = sysconf(_SC_PAGESIZE);                                       //  system page size
   ncpus = sysconf(_SC_NPROCESSORS_ONLN);                                  //  SMP CPU count

   prio = getpriority(PRIO_PROCESS,0);                                     //  current priority
   if (prio > -2 && getuid() == 0) {
      err = setpriority(PRIO_PROCESS,0,-2);                                //  boost priority if allowed
      if (err) printf("watsup: unable to boost priority \n");
      else printf("watsup: priority set to -2 \n");
   }

   setreportfonts();                                                       //  setup report fonts from font size

   m_start();                                                              //  start report running
   return 0;
}


//  window draw event - recalibrate report window dimensions

int m_draw(GtkWidget *, cairo_t *cr)                                       //  gtk3
{
   GdkWindow   *textwin;

   gtk_window_get_position(GTK_WINDOW(Pwin),&pxpos,&pypos);                //  track current size and position
   gtk_window_get_size(GTK_WINDOW(Pwin),&pww,&phh);

   textwin = gtk_text_view_get_window(GTK_TEXT_VIEW(mLog),GTK_TEXT_WINDOW_WIDGET);
   phh2 = gdk_window_get_height(textwin);                                  //  text window height

   save_config();                                                          //  save configuration parameters
   return 0;
}


//  mouse click event - get row clicked, get process data from current report

void m_mouse(GtkWidget *, GdkEventButton *event, void *)                   //  v.1.2
{
   int         mpy, row, ii;
   int         kill, killpid;                                              //  process to kill, PID
   char        killfname[50];                                              //  executable filename

   gdk_window_set_cursor(gdkwin,cursor);                                   //  this must be done continuously else
                                                                           //    pointer reverts to text-edit  v.2.4
   if (event->type != GDK_BUTTON_RELEASE) return;

   mpy = int(event->y);                                                    //  mouse y-position
   
   row = mpy / fontheight - sysreplines + 2;                               //  report row selected
   if (row <= 0) return;
   
   for (ii = 0; ii < nrec3; ii++)
      if (prec3[ii].row == row) break;
   if (ii == nrec3) return;

   killpid = prec3[ii].pid;
   if (killpid <= 0) return;

   strncpy0(killfname,prec3[ii].exefname,49);

   kill = zmessageYN(Pwin,"Kill this process? \n %d  %s",killpid,killfname);
   if (kill) shell_ack("kill -KILL %d",killpid);

   return;
}


//  menu selection events

void menufunc(GtkWidget *item, const char *menu)
{
   if (strEqu(menu,"parms")) m_parms();
   if (strEqu(menu,"root")) m_getroot();
   if (strEqu(menu,"start")) m_start();
   if (strEqu(menu,"stop")) m_stop();
   if (strEqu(menu,"sort")) m_sort();
   if (strEqu(menu,"save")) m_save();
   if (strEqu(menu,"print")) m_print();
   if (strEqu(menu,"quit")) m_quit();
   if (strEqu(menu,"help")) m_help();
   return;
}


//  edit report parameters

void m_parms() 
{ 
   int   zd_parms_event(zdialog *zd, cchar *event);                        //  dialog event function

   zdialog     *zd;
   
   zd = zdialog_new("edit watsup parameters",Pwin,"done","cancel",null);
   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog | space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog | space=5");
   zdialog_add_widget(zd,"label","lab1","vb1","sample period");
   zdialog_add_widget(zd,"label","lab2","vb1","font size");
   zdialog_add_widget(zd,"label","lab3","vb1","match strings");
   zdialog_add_widget(zd,"spin","period","vb2","0.1|99.0|0.1|1.0");
   zdialog_add_widget(zd,"spin","font","vb2","5|20|1|9");
   zdialog_add_widget(zd,"comboE","match","vb2",procmatch);
   
   zdialog_stuff(zd,"period",Pinterval);                                   //  stuff widgets with current params
   zdialog_stuff(zd,"font",fontsize);
   
   for (int ii = 0; ii < Nmatchhist; ii++)                                 //  stuff match strings history     v.2.6
      zdialog_cb_app(zd,"match",matchhist[ii]);

   zdialog_run(zd,zd_parms_event);                                         //  run dialog
   return;
}


int zd_parms_event(zdialog *zd, cchar *event)                              //  dialog event function
{
   const char  *pp;
   int         ii;
   
   if (! zd->zstat) return 0;
   
   if (zd->zstat != 1) {
      zdialog_free(zd);
      return 0;
   }

   zdialog_fetch(zd,"period",Pinterval);                                   //  get new params from widgets
   zdialog_fetch(zd,"font",fontsize);
   zdialog_fetch(zd,"match",procmatch,99);
   
   setreportfonts();                                                       //  calculate font metrics
   
   if (*procmatch) {
      for (int ii = 0; ii < 10; ii++) {                                    //  get process match strings  v.1.3
         pp = strField(procmatch,' ',ii+1);
         if (pp) matchstrings[ii] = zstrdup(pp);                           //  v.1.4.1
         else matchstrings[ii] = 0;
      }
   }
   else matchstrings[0] = 0;
   if (! matchstrings[0]) *procmatch = 0;                                  //  only blank(s)
   
   for (ii = 0; ii < 20; ii++) {
      pp = zdialog_cb_get(zd,"match",ii);                                  //  capture combo box entries
      if (! pp) break;                                                     //    >> matchhist[] list        v.2.6
      strncpy0(matchhist[ii],pp,30);
   }
   Nmatchhist = ii;

   zdialog_free(zd);
   return 1;
}


//  set up report fonts from user font size setting, recalculate font metrics

void setreportfonts()
{
   PangoLanguage           *plang;
   PangoFontDescription    *pfontdesc;
   PangoContext            *pcontext;
   PangoFont               *pfont;
   PangoFontMetrics        *pmetrics;
   int                     fontascent, fontdescent;
   
   snprintf(headerfont,19,"Monospace bold %d",fontsize);                   //  report header font, bold
   snprintf(bodyfont,19,"Monospace %d",fontsize);                          //  report body font

   pfontdesc = pango_font_description_from_string(bodyfont);               //  set default font for report window
   gtk_widget_override_font(mLog,pfontdesc);

   plang = pango_language_from_string("en-US");                            //  get font metrics (what a mess)
   pcontext = gtk_widget_get_pango_context(mLog);
   pfont = pango_context_load_font(pcontext,pfontdesc);
   pmetrics = pango_font_get_metrics(pfont,plang);
   fontascent = pango_font_metrics_get_ascent(pmetrics) / PANGO_SCALE;
   fontdescent = pango_font_metrics_get_descent(pmetrics) / PANGO_SCALE;
   fontheight = fontascent + fontdescent;                                  //  effective report line height

   return;
}


//  restart as root if password OK

void m_getroot()
{
   save_config();                                                          //  save configuration parameters
   beroot(m_argc-1,m_argv+1);                                              //  does not return     v.2.9
}


//  start reporting

void m_start()
{
   if (Pruns) return;
   Pruns = 1;
   proc_report(0);
   return;
}


//  stop reporting

void m_stop()
{
   Pruns = 0;
   return;
}


//  re-sort top processes to top of list

void m_sort()
{ 
   Fresort = 1;                                                            //  set flag for process report
   return;
}


//  save report to a file

void m_save()
{ 
   wfilesave(mLog);
   return;
}


//  print report on default printer

void m_print()
{ 
   wprintp(mLog);
   return;
}


//  exit process detail report

void m_quit()
{
   Pruns = 0;
   gtk_widget_destroy(Pwin);                                               //  kill window
   save_config();                                                          //  save configuration parameters
   return;
}


//  display help file (man page)

void m_help()                                                              //  menu function
{
   popup_command("man watsup",700,500);                                    //  show man page
   return;
}


//  process report main function

int proc_report(void *)
{
   int      millisecs = Pinterval * 1000;                                  //  sample interval, milliseconds

   if (! Pruns) return 0;                                                  //  report stopped

   time2 = get_seconds();

   if (! time1) {                                                          //  first call, no time #1
      time1 = time2;
      get_sys_stats(&srec1);                                               //  get system stats #1
      get_proc_stats(prec1,nrec1);                                         //  get process stats #1
      g_timeout_add(millisecs,proc_report,0);                              //  reschedule and exit
      return 0;
   }
   
   get_sys_stats(&srec2);                                                  //  get system stats #2
   get_proc_stats(prec2,nrec2);                                            //  get process stats #2

   elapsed = time2 - time1;
   report_sys_stats();                                                     //  system report for interval #1 to #2
   report_proc_stats();                                                    //  process report for interval #1 to #2

   time1 = time2;                                                          //  copy time #2 to time #1
   copy_sys_stats();                                                       //  copy system stats #2 to #1
   copy_proc_stats();                                                      //  copy process stats #2 to #1
   
   g_timeout_add(millisecs,proc_report,0);                                 //  reschedule and exit
   return 0;
}


//  get system level statistics from various /proc/xxxx files

void get_sys_stats(sysrec *srec)
{
   int         ii;
   FILE        *fid;
   char        *pp, buff[1000];
   const char  *ppc;
   double      val1, val2, val3, val4;
   
   memset(srec,0,sizeof(sysrec));                                          //  zero data
   
   fid = fopen("/proc/stat","r");                                          //  read /proc/stat
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         if (strncmp(pp,"cpu",3) == 0) {                                   //  look for "cpuN" record
            ii = atoi(pp+3);
            if (ii < 0 || ii >= ncpus) continue;
            parseprocrec(pp,2,&val1,4,&val2,null);                         //  get user and system mode jiffies
            srec->cpubusy += val1 + val2;                                  //  v.1.8
            if (ii == ncpus-1) break;
         }
      }
      fclose(fid);
   }

   parseprocfile("/proc/meminfo","MemTotal:",&val1,"MemFree:",&val2,       //  memory statistics
                                 "Cached:",&val3,null);
   srec->memused = val1 - val2 - val3;                                     //  used = total-free-cached   v.1.5
   srec->memfree = val2;
   srec->memcache = val3;

   parseprocfile("/proc/vmstat","pgfault",&val1,"pgmajfault",&val2,null);  //  page fault statistics
   srec->pfsoft = val1;
   srec->pfhard = val2;

   fid = fopen("/proc/diskstats","r");                                     //  read disk statistics
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         ppc = strField(buff,' ',1);                                       //  screen for "8"      v.1.4
         if (! ppc) continue;
         if (strNeq(ppc,"8")) continue;
         ppc = strField(buff,' ',3);                                       //  disk, "sda" or "sda1" etc.
         if (! ppc) continue;
         if (ppc[3]) continue;                                             //  ignore partitions
         parseprocrec(pp,4,&val1,6,&val2,8,&val3,10,&val4,null);
         srec->diskops += val1 + val3;                                     //  read/write counts and sectors
         srec->disksecs += val2 + val4;
      }
      fclose(fid);
   }
   
   fid = fopen("/proc/net/dev","r");                                       //  read network statistics
   if (fid) {
      while ((pp = fgets(buff,999,fid)))
      {
         pp = strchr(buff,':');                                            //  look for device "eth0:" etc.
         if (! pp) continue;
         parseprocrec(pp+1,1,&val1,9,&val2,null);                          //  get bytes received and transmitted
         srec->netrecv += val1;
         srec->netxmit += val2;
      }
      fclose(fid);
   }
   
   return;   
}


//  get process level statistics from /proc/<pid>/stat files
      
void get_proc_stats(pidrec *prec, int &nnp)
{
   DIR            *procdir;
   FILE           *fid;
   struct dirent  *procent;
   char           *ppid, *pp, *pp2;
   char           username[20], exefname[40];
   char           buff1[100], buff2[1000];
   int            ii, cc, pid, fkeep;
   double         uid, cpu1, cpu2, cpu3, cpu4;
   double         mem, pfs1, pfs2, rbs, wbs;

   procdir = opendir("/proc");
   if (! procdir) zappcrash("open /proc failed: %s",strerror(errno));
   
   nnp = 0;

   while (true)                                                            //  loop per PID
   {
      procent = readdir(procdir);                                          //  read /proc directory
      if (! procent) break;
      ppid = procent->d_name;
      if (*ppid < '1' || *ppid > '9') continue;                            //  skip non-pid entries
      
      pid = atoi(ppid);
      pfs1 = pfs2 = cpu1 = cpu2 = cpu3 = cpu4 = mem = 0;
      *exefname = *username = 0;
      
      snprintf(buff1,99,"/proc/%s/status",ppid);                           //  read file /proc/<pid>/status
      parseprocfile(buff1,"Uid:",&uid,null);                               //  get real UID     v.2.0

      snprintf(buff1,99,"/proc/%s/environ",ppid);                          //  read file /proc/<pid>/environ
      fid = fopen(buff1,"r");
      if (! fid) continue;
      pp = fgets(buff2,999,fid);
      fclose(fid);
      if (! pp) continue;
      
      if (uid == 0) strcpy(username,"root");                               //  uid 0, user = root     v.2.0
      
      else {
         while (*pp) {
            pp2 = strstr(pp,"USER");                                       //  look for xxxUSERxxx=name
            if (pp2) {                                                     //  v.2.0
               pp2 = strchr(pp2+4,'=');
               if (pp2) {
                  strncpy0(username,pp2+1,19);                             //  found, use it
                  break;
               }
            }
            pp += strlen(pp) + 1;
         }
         if (! *pp) strcpy(username,"none");                               //  v.2.0
      }

      snprintf(buff1,99,"/proc/%s/exe",ppid);                              //  proc/<pid>/exe
      cc = readlink(buff1,buff2,999);                                      //  executable pathname/filename
      if (cc <= 0) continue;
      buff2[cc] = 0;
      pp = (char *) strrchr(buff2,'/');                                    //  isolate filename
      if (pp) pp++;
      else pp = buff2;
      if (strstr(proc_ignore,pp)) continue;                                //  ignore some processes
      strncpy0(exefname,pp,39);                                            //  keep filename only
      
      if (matchstrings[0]) {
         fkeep = 0;                                                        //  match pid, user, exefile
         for (ii = 0; ii < 10; ii++) {                                     //    with user strings     v.1.3
            if (! matchstrings[ii]) break;
            if (MatchWild(matchstrings[ii],ppid) == 0) fkeep = 1;
            else if (MatchWild(matchstrings[ii],username) == 0) fkeep = 1;
            else if (MatchWild(matchstrings[ii],exefname) == 0) fkeep = 1;
            if (fkeep) break;
         }
         if (! fkeep) continue;
      }

      snprintf(buff1,99,"/proc/%s/stat",ppid);                             //  read file /proc/<pid>/stat
      fid = fopen(buff1,"r");
      if (! fid) continue;
      pp = fgets(buff2,999,fid);
      fclose(fid);
      if (! pp) continue;
      
      pp = strchr(pp,')');                                                 //  closing ')' after (short) filename
      if (pp) parseprocrec(pp+1, 10,&pfs1, 11,&pfs2, 12,&cpu1,             //  parse and get wanted fields
                       13,&cpu2, 14,&cpu3, 15,&cpu4, 22,&mem, null);
      
      snprintf(buff1,99,"/proc/%s/io",ppid);                               //  read file /proc/<pid>/io
      parseprocfile(buff1,"read_bytes:",&rbs,"write_bytes:",&wbs,null);    //  get physical I/O byte counts  v.1.6

      prec[nnp].pid = pid;                                                 //  build proc record
      prec[nnp].cpu = cpu1 + cpu2 + cpu3 + cpu4;                           //  user and system jiffies  v.1.5
      prec[nnp].mem = mem;                                                 //  real memory
      prec[nnp].pfs = pfs1 + pfs2;                                         //  major page faults   v.1.5
      prec[nnp].dmbs = (rbs + wbs);                                        //  I/O bytes 
      strcpy(prec[nnp].username,username);                                 //  user name
      strcpy(prec[nnp].exefname,exefname);                                 //  executable file
      
      if (++nnp == maxpid) zappcrash("exceeded %d processes",maxpid);
   }

   closedir(procdir);
   return;
}


//  create report from two sets of system data and time interval

void report_sys_stats()
{
   int         ii, jj, lines = 4;                                          //  v.1.8
   double      cpu1, cpupct, MBused, MBfree, MBcache;
   double      pfsoft, pfhard;
   double      diskops, diskMBs;
   double      netrecv, netxmit;
   char        *pp, sysline[4][100];
   const char  *header = " period    CPU    memory MB     PFs/sec      disk/sec    net KB/sec \n";

   repline = 1;
   wprintx(mLog,repline++,"\n");
   wprintx(mLog,repline++,header,headerfont);

   cpu1 = srec2.cpubusy - srec1.cpubusy;                                   //  CPU busy time   v.1.8
   cpupct = cpu1 * 100.0 * jiffy / elapsed;

   MBused = srec2.memused / 1024.0;                                        //  get used, free, cached memory, MB
   MBfree = srec2.memfree / 1024.0;
   MBcache = srec2.memcache / 1024.0;
   
   pfsoft = (srec2.pfsoft - srec1.pfsoft) / elapsed;                       //  hard and soft faults / sec
   pfhard = (srec2.pfhard - srec1.pfhard) / elapsed;
   
   diskops = (srec2.diskops - srec1.diskops) / elapsed;                    //  disk IOs and KB / sec
   diskMBs = (srec2.disksecs - srec1.disksecs) / elapsed;
   diskMBs = diskMBs / 2048;                                               //  512 byte sectors to megabytes

   netrecv = (srec2.netrecv - srec1.netrecv) / elapsed / 1024;             //  network KB / sec
   netxmit = (srec2.netxmit - srec1.netxmit) / elapsed / 1024;

   memset(sysline,' ',sizeof(sysline));                                    //  clear report lines

   for (ii = 0; ii < lines; ii++)
   {
      pp = sysline[ii];                                                    //  pointer to line 0 ... 3
      
      if (ii == 0) sprintf(pp,"%7.2f",elapsed);                            //  elapsed time on line 0
      
      if (ii == 0) sprintf(pp+10,"%4.0f",cpupct);                          //  CPU % on line 0

      if (ii == 0) sprintf(pp+17," used: %.0f",MBused);                    //  memory statistics on lines 0-2
      if (ii == 1) sprintf(pp+17,"cache: %.0f",MBcache);
      if (ii == 2) sprintf(pp+17,"avail: %.0f",MBfree);
      
      if (ii == 0) sprintf(pp+32,"soft: %.0f",pfsoft);                     //  page fault stats on lines 0-1
      if (ii == 1) sprintf(pp+32,"hard: %.0f",pfhard);
      
      if (ii == 0) sprintf(pp+45,"IOs: %.0f",diskops);                     //  disk I/O stats on lines 0-1
      if (ii == 1) sprintf(pp+45,"MBs: %.0f",diskMBs);
      
      if (ii == 0) sprintf(pp+57,"recv: %.0f",netrecv);                    //  network stats on lines 0-1
      if (ii == 1) sprintf(pp+57,"xmit: %.0f",netxmit);
   }
   
   for (ii = 0; ii < lines; ii++)                                          //  output the lines
   {
      pp = sysline[ii];
      for (jj = 0; jj < 99; jj++) if (pp[jj] < ' ') pp[jj] = ' ';          //  remove imbedded nulls
      for (jj = 98; jj; jj--) if (pp[jj] > ' ') break;                     //  and trailing blanks
      pp[jj+1] = 0;
      wprintf(mLog,repline++,"%s \n",pp);
   }
}


//  create report from two sets of process data and time interval

void report_proc_stats()
{
   const char  *header1 = "                  real    hard    I/O   \n";
   const char  *header2 = "   PID     CPU     MBs    PF/s    MB/s   User     Executable \n";

   double      cpupct, MB, dmbs, weight;
   char        scpupct[8];
   int         row, nrep;
   int         rowmap[maxpid];
                                                                           //  sort keys:
   int      key1[2][3] = { { 8, 8, 8 }, { 0, 4, 3 } };                     //    weight descending, PID ascending
   int      key2[1][3] = { { 4, 4, 3 } };                                  //    row ascending
   int      key3[1][3] = { { 0, 4, 3 } };                                  //    PID ascending

   wprintx(mLog,repline++,header1,headerfont);                             //  output headers
   wprintx(mLog,repline++,header2,headerfont);
   
   sysreplines = repline;                                                  //  lines used before 1st process line

   for (rec2 = 0; rec2 < nrec2; rec2++)                                    //  copy proc recs #2 to #3 for reporting
      prec3[rec2] = prec2[rec2];
   nrec3 = nrec2;

   for (rec1 = rec3 = 0; rec3 < nrec3; rec1++, rec3++)                     //  loop through proc recs #3
   {
      if (rec1 >= nrec1 || prec1[rec1].pid != prec3[rec3].pid) {           //  find same PID in proc recs #1
         for (rec1 = 0; rec1 < nrec1; rec1++)                              //  (utilize order, but don't count on it)
            if (prec1[rec1].pid == prec3[rec3].pid) break;
      }
      
      if (rec1 < nrec1) {                                                  //  if found, calculate deltas:
         prec3[rec3].cpu -= prec1[rec1].cpu;                               //    cpu jiffies
         prec3[rec3].pfs -= prec1[rec1].pfs;                               //    hard page faults
         prec3[rec3].dmbs -= prec1[rec1].dmbs;                             //    I/O bytes

         prec3[rec3].weight = prec1[rec1].weight;                          //  copy weight and row from proc rec #1
         prec3[rec3].row = prec1[rec1].row;                                //    to proc rec #3
      }

      else {
         prec3[rec3].weight = 0;                                           //  not found (new process was started)
         prec3[rec3].row = 0;                                              //  weight and row = 0
      }
   }

   for (rec3 = 0; rec3 < nrec3; rec3++)                                    //  calc. new sort weights
   {
      weight = prec3[rec3].cpu + 0.1 * prec3[rec3].pfs + prec3[rec3].dmbs;
      if (prec3[rec3].row) weight += 0.1;
      weight = 0.25 * weight + 0.75 * prec3[rec3].weight;                  //  use historical weighted average
      prec3[rec3].weight = weight;
   }
      
   MemSort((char *) prec3, sizeof(pidrec), nrec3, key1, 2);                //  sort proc recs by weight descending

   topmost = phh2 / fontheight - repline;                                  //  proc recs fitting in report window
   if (topmost < 3) topmost = 3;
   if (topmost < nrec3) nrep = topmost;                                    //  topmost proc recs to report
   else nrep = nrec3;

   for (row = 0; row < nrep; row++) rowmap[row] = 0;                       //  clear row map

   for (rec3 = 0; rec3 < nrep; rec3++)                                     //  assign report row to topmost proc recs
   {
      row = prec3[rec3].row;                                               //  keep prior report row
      if (row > 0 && row < nrep) rowmap[row-1] = 1;                        //    if available and in range
      else prec3[rec3].row = 0;
   }
   
   for (rec3 = row = 0; rec3 < nrep; rec3++)                               //  assign rest from top down
   {
      if (prec3[rec3].row > 0) continue;
      while (rowmap[row] > 0) row++;
      prec3[rec3].row = 1 + row++;
   }

   for (rec3 = nrep; rec3 < nrec3; rec3++)                                 //  clear all rows after topmost
      prec3[rec3].row = 0;
   
   if (Fresort) {                                                          //  if a re-sort was requested - 
      for (rec3 = 0; rec3 < nrep; rec3++)                                  //    keep current order by weight
            prec3[rec3].row = rec3 + 1;
      Fresort = 0;
   }

   MemSort((char *) prec3, sizeof(pidrec), nrep, key2, 1);                 //  sort topmost proc recs by row

   for (rec3 = 0; rec3 < nrep; rec3++)                                     //  output each proc record
   {
      cpupct = prec3[rec3].cpu * 100.0 * jiffy / elapsed;                  //  format 0.0 to 9.9, then 10 to 100
      if (cpupct > 101 * ncpus) cpupct = 0;                                //  terminated child added to parent
      if (cpupct < 9.95) sprintf(scpupct,"%4.1f",cpupct);                  //  (stop double-counts)
      else sprintf(scpupct,"%4.0f",cpupct);

      MB = prec3[rec3].mem * pagesize / mega;
      dmbs = prec3[rec3].dmbs / mega / elapsed;                            //  bugfix  v.1.6

      wprintf(mLog,repline++,"%6d  %6s  %6.1f  %6.0f  %6.1f   %-8s %s \n", 
               prec3[rec3].pid, scpupct, MB, prec3[rec3].pfs, 
               dmbs, prec3[rec3].username, prec3[rec3].exefname);
   }

   wclear(mLog,repline);                                                   //  clear rest of window
   
   MemSort((char *) prec2, sizeof(pidrec), nrec2, key3, 1);                //  sort proc recs #2 by PID ascending
   MemSort((char *) prec3, sizeof(pidrec), nrec3, key3, 1);                //  sort proc recs #3 by PID ascending
   
   for (rec2 = 0; rec2 < nrec2; rec2++)                                    //  copy new weights and rows
   {                                                                       //    to proc recs #2
      if (prec2[rec2].pid != prec3[rec2].pid) zappcrash("bug 23");
      prec2[rec2].weight = prec3[rec2].weight;
      prec2[rec2].row = prec3[rec2].row;
   }
   
   return;
}


//  copy system stats data set #2 to data set #1

void copy_sys_stats()
{
   srec1 = srec2;
}


//  copy proc stats data set #2 to data set #1

void copy_proc_stats()
{
   for (rec2 = 0; rec2 < nrec2; rec2++)
         prec1[rec2] = prec2[rec2];
   nrec1 = nrec2;
   return;
}


//  load and save configuration parameters                                 //  v.2.0

void load_config()
{
   int      ii, nn, mm1, mm2;
   FILE     *fid = 0;

   snprintf(configfile,199,"%s/config",get_zuserdir());                    //  get /home/<user>/.watsup/config

   fid = fopen(configfile,"r");
   if (! fid) return;

   nn = fscanf(fid,"twin position: %d %d ",&mm1,&mm2);                     //  last tiny window position
   if (nn == 2) {
      txpos = mm1;
      typos = mm2;
   }

   nn = fscanf(fid,"twin size: %d %d ",&mm1,&mm2);                         //  last tiny window size
   if (nn == 2) {
      tww = mm1;
      thh = mm2;
   }

   nn = fscanf(fid,"pwin position: %d %d ",&mm1,&mm2);                     //  last process window position
   if (nn == 2) {
      pxpos = mm1;
      pypos = mm2;
   }

   nn = fscanf(fid,"pwin size: %d %d ",&mm1,&mm2);                         //  last process window size
   if (nn == 2) {
      pww = mm1;
      phh = mm2;
   }
   
   for (ii = 0; ii < 20; ii++) {                                           //  last match string history       v.2.6
      nn = fscanf(fid,"match: %30c ",matchhist[ii]);                       //  30 chars. + \n
      if (nn != 1) break;
      matchhist[ii][29] = 0;
      strTrim(matchhist[ii]);
   }
   Nmatchhist = ii;

   fclose(fid);
   
   return;
}

void save_config()
{
   FILE     *fid = 0;

   fid = fopen(configfile,"w");
   if (! fid) return;

   fprintf(fid,"twin position: %d %d \n",txpos,typos);                     //  to use for next time
   fprintf(fid,"twin size: %d %d \n",tww,thh);
   fprintf(fid,"pwin position: %d %d \n",pxpos,pypos);
   fprintf(fid,"pwin size: %d %d \n",pww,phh);
   
   for (int ii = 0; ii < Nmatchhist; ii++)                                 //  v.2.6
      fprintf(fid,"match: %-30s \n",matchhist[ii]);

   fprintf(fid,"\n");
   fclose(fid);
   return;
}


//  supply unused zdialog callback function

void KBstate(GdkEventKey *event, int state) 
{ return; }





